import json
import logging
import os
from tempfile import TemporaryFile

from crypto.models.upload_file import DSResponse_UploadFile
from django.conf import settings
from django.core.files import File
from django.db import transaction
from isc_common.common import SFTP, FILES_STORE_UNKNOWN_TYPE
from isc_common.common.UploadItem import UploadItem
from isc_common.dropzone import Dz
from isc_common.fields.files import make_REPLACE_FILE_PATH
from isc_common.http.DSResponse import JsonResponseWithException
from isc_common.models.image_types import Image_types
from isc_common.models.images import Images
from websocket import create_connection

from lfl_admin.common.management.commands import delete_img_refs

logger = logging.getLogger(__name__)


class Common_UploadImage(DSResponse_UploadFile):
    def upload_image(self, request):

        self.dictionary = dict(id=request.GET.get('id'))

        self.dz_dictionary = Dz(request.POST)

        self.host = request.GET.get('host')
        self.port = request.GET.get('port')
        self.image_type_name = request.GET.get('image_type_name')
        self.keyImage = request.GET.get('keyImage')

        if self.image_type_name is None:
            raise Exception('Не задан параметр image_type_name.')
        if not self.port:
            self.port = 80
        self.channel = request.GET.get('ws_channel')
        self.file = request.FILES.get('upload_attatch')

        self.dictionary.update(dict(
            real_file_name=self.file.name,
            stored_file_name=self.dz_dictionary.dzuuid,
            file_size=int(self.dz_dictionary.dztotalfilesize),
            file_mime_type=self.file.content_type)
        )

        item = UploadItem(dictionary=self.dictionary)

        def load_str(pers):
            return f'Загружено: {pers} %'

        image_type, _ = Image_types.objects.get_or_create(code=self.image_type_name, keyimage=self.keyImage, defaults=dict(name=self.image_type_name))

        res = Images.objects.getOptional(
            size=item.file_size,
            real_name=item.real_file_name,
        )

        if res is not None:
            attfile = make_REPLACE_FILE_PATH(str(res.attfile))

            if isinstance(settings.FILES_STORE, str):
                ex = os.path.exists(attfile)
                if ex is False:
                    delete_img_refs(old_id=res.id)
                    res = None
            else:
                ssh_clients = settings.SSH_CLIENTS.client(settings.FILES_STORE)
                ex = ssh_clients.exists(attfile)
                if ex is False:
                    delete_img_refs(old_id=res.id)
                    res = None

        if res is None:
            with TemporaryFile() as src:
                src.seek(int(self.dz_dictionary.dzchunkbyteoffset))
                src.write(self.file.read())

                pers = round((int(self.dz_dictionary.dzchunkindex) * 100) / int(self.dz_dictionary.dztotalchunkcount), 2)
                if self.last_chunk:
                    pers = 100

                logger.debug(f'{load_str(pers)}, шаг : {int(self.dz_dictionary.dzchunkindex) + 1} из {self.dz_dictionary.dztotalchunkcount}')

                if self.last_chunk:
                    logger.debug(f'Загружен файл: {item.real_file_name}.')
                    logger.debug(f'Запись временного файла.')

                    with open(item.full_path, 'w+b') as destination:
                        src.seek(0)
                        destination.write(src.read())
                        logger.debug(f'Запись временного файла выполнена.')

                    logger.debug(f'Запись: {item.full_path}.')
                    with open(item.full_path, 'rb') as src:
                        src.seek(0)
                        fileObj = File(src)

                        if isinstance(settings.FILES_STORE, str):
                            file_store = os.path.abspath(settings.FILES_STORE)
                        elif isinstance(settings.FILES_STORE, dict) and settings.FILES_STORE.get('mode') == SFTP:
                            file_store = settings.FILES_STORE.get('PATH')
                        else:
                            raise Exception(FILES_STORE_UNKNOWN_TYPE)

                        res = Images.objects.create(
                            attfile=fileObj,
                            file_store=file_store,
                            format=item.file_format,
                            mime_type=item.file_mime_type,
                            size=item.file_size,
                            real_name=item.real_file_name,
                        )

                        ws = create_connection(f"ws://{self.host}:{self.port}/ws/{self.channel}/")
                        ws.send(json.dumps(dict(id=item.id, item_id=res.id, type="uploaded")))
                        ws.close()

                    logger.debug(f'Запись: {item.real_file_name} ({fileObj.name}) завершена.')

                    self.remove(item.full_path)
                    logger.debug(f'Удаление: {item.full_path} завершено.')

        return res, image_type, item.id

    @property
    def last_chunk(self):
        res = int(self.dz_dictionary.dztotalchunkcount) == int(self.dz_dictionary.dzchunkindex) + 1
        return res


@JsonResponseWithException()
class DSResponse_UploadImage(Common_UploadImage):
    def __init__(self, request):
        self.upload_image(request=request)


@JsonResponseWithException()
class DSResponse_CommonUploadImage(Common_UploadImage):
    def __init__(self, request, model, image_model, field_main_model=''):
        with transaction.atomic():
            file = request.FILES.get('upload_attatch')
            if file is not None:
                image, type, id = self.upload_image(request=request)

                main_model = eval(f"model.objects.get( id={id} ){f'.{field_main_model}' if field_main_model else ''}")
                if main_model is None:
                    raise Exception(f'Для field_main_model {field_main_model} не найдено main_model')

                main_model_image = image_model.objects.getOptional(main_model=main_model, image=image, type=type, deleted_at__isnull=False)

                if main_model_image is not None:
                    main_model_image.soft_restore()
                    main_model_image.save()
                else:
                    main_model_image = image_model.objects.create(main_model=main_model, type=type, image=image)

                image_model.objects.filter(main_model=main_model, type=type, deleted_at__isnull=True).exclude(id=main_model_image.id).soft_delete()

                logger.debug(f'Created: {main_model_image}')



from isc_common.http.DSResponse import DSResponseUpdate, DSResponseAdd, DSResponse, JsonResponseWithException
from isc_common.http.RPCResponse import RPCResponseConstant
from isc_common.http.response import JsonResponse
from isc_common.models.text_informations import Text_informations, Text_informationsManager


@JsonResponseWithException()
def Text_informations_Fetch(request):
    return JsonResponse(
        DSResponse(
            request=request,
            data=Text_informations.objects.
                select_related().
                get_range_rows1(
                request=request,
                function=Text_informationsManager.getRecord
            ),
            status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Text_informations_Add(request):
    return JsonResponse(DSResponseAdd(data=Text_informations.objects.createFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Text_informations_Update(request):
    return JsonResponse(DSResponseUpdate(data=Text_informations.objects.updateFromRequest(request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Text_informations_Remove(request):
    return JsonResponse(DSResponse(request=request, data=Text_informations.objects.deleteFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Text_informations_Lookup(request):
    return JsonResponse(DSResponse(request=request, data=Text_informations.objects.lookupFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Text_informations_Info(request):
    return JsonResponse(DSResponse(request=request, data=Text_informations.objects.get_queryset().get_info(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Text_informations_Copy(request):
    return JsonResponse(DSResponse(request=request, data=Text_informations.objects.copyFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)



import logging

from isc_common.models.audit import AuditModel, AuditManager, AuditQuerySet
from isc_common.models.base_ref import BaseRef, BaseRefManager, BaseRefQuerySet

logger = logging.getLogger(__name__)


class Fragments_item_typesQuerySet(BaseRefQuerySet):
    pass


class Fragments_item_typesManager(BaseRefManager):

    @classmethod
    def getRecord(cls, record ) :
        res = {
            'id': record.id,
            'code': record.code,
            'name': record.name,
            'description': record.description,
            'editing': record.editing,
            'deliting': record.deliting,
        }
        return res

    def get_queryset(self):
        return Fragments_item_typesQuerySet(self.model, using=self._db)


class Fragments_item_types(BaseRef):
    objects = Fragments_item_typesManager()

    def __str__(self):
        return f'ID:{self.id}, code: {self.code}, name: {self.name}, description: {self.description}'

    def __repr__(self):
        return self.__str__()

    class Meta:
        verbose_name = 'Типы элементов фрагментов'



# Generated by Django 3.2.7 on 2021-09-05 17:52

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('isc_common', '0128_images_file_name'),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name='images',
            unique_together={('file_name', 'size')},
        ),
    ]



import logging

from isc_common.auth.models.user import User
from isc_common.fields.related import ForeignKeyProtect
from isc_common.models.audit import AuditManager, AuditQuerySet, AuditModel

logger = logging.getLogger(__name__)


class AuditModelExQuerySet(AuditQuerySet):
    pass


class AuditModelExManager(AuditManager):

    @classmethod
    def getRecord(cls, record):
        res = {
            'id': record.id,
            'editing': record.editing,
            'deliting': record.deliting,
        }
        return res

    def get_queryset(self):
        return AuditModelExQuerySet(self.model, using=self._db)


class AuditModelEx(AuditModel):
    creator = ForeignKeyProtect(User, related_name='+', null=True, blank=True)
    editor = ForeignKeyProtect(User, related_name='+', null=True, blank=True)
    objects = AuditModelExManager()

    def __str__(self):
        return f'ID:{self.id}'

    def __repr__(self):
        return self.__str__()

    class Meta:
        verbose_name = ' Расширенная AuditModel'
        abstract = True



# Generated by Django 3.2.3 on 2021-05-14 15:25

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('tracker', '0088_alter_messages_files_real_name'),
    ]

    operations = [
        migrations.AlterField(
            model_name='messages_files',
            name='real_name',
            field=models.TextField(db_index=True, unique=True, verbose_name='Первоначальное имя файла'),
        ),
    ]



from isc_common.http.DSResponse import DSResponseUpdate, DSResponseAdd, DSResponse, JsonResponseWithException
from isc_common.http.RPCResponse import RPCResponseConstant
from isc_common.http.response import JsonResponse
from react.models.fragment_params import Fragment_params
from react.models.fragment_params_upload_image import DSResponse_Fragment_params_UploadImage
from react.models.fragment_params_view import Fragment_params_view, Fragment_params_viewManager


@JsonResponseWithException()
def Fragment_params_Fetch(request):
    return JsonResponse(
        DSResponse(
            request=request,
            data=Fragment_params_view.objects.
                select_related().
                get_range_rows1(
                request=request,
                function=Fragment_params_viewManager.getRecord
            ),
            status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Fragment_params_Add(request):
    return JsonResponse(DSResponseAdd(data=Fragment_params.objects.createFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Fragment_params_Update(request):
    return JsonResponse(DSResponseUpdate(data=Fragment_params.objects.updateFromRequest(request,propsArr=['use_generator']), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Fragment_params_Remove(request):
    return JsonResponse(DSResponse(request=request, data=Fragment_params.objects.deleteFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Fragment_params_Lookup(request):
    return JsonResponse(DSResponse(request=request, data=Fragment_params.objects.lookupFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Fragment_params_Info(request):
    return JsonResponse(DSResponse(request=request, data=Fragment_params.objects.get_queryset().get_info(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Fragment_params_Copy(request):
    return JsonResponse(DSResponse(request=request, data=Fragment_params.objects.copyFromRequest(request=request), status=RPCResponseConstant.statusSuccess).response)


@JsonResponseWithException()
def Fragment_params_ImagesUpload(request):
    DSResponse_Fragment_params_UploadImage(request)
    return JsonResponse(dict(status=RPCResponseConstant.statusSuccess))



import os
import uuid
from pathlib import Path

import magic
from django.conf import settings

from isc_common.common.UploadItem import UploadItem


class UploadItemEx(UploadItem):

    def get_path(self, name):
        return os.path.dirname(os.path.abspath(name)).replace(os.sep, os.altsep) if os.altsep else os.path.dirname(os.path.abspath(name))

    def __init__(self, logger=None, **kwargs):
        super().__init__(logger=logger, dictionary=kwargs)

        stat = None
        if self.real_file_name is not None and self.file_mime_type is None:
            _, self.file_name = os.path.split(self.real_file_name)
            mime = magic.Magic(mime=True)
            if isinstance(settings.OLD_FILES, dict):
                o_ssh_client = settings.SSH_CLIENTS.client(settings.OLD_FILES)
                if o_ssh_client.exists(self.real_file_name):
                    self.file_mime_type = mime.from_buffer(o_ssh_client.open(self.real_file_name, 'rb').read())
                    stat = o_ssh_client.stat(self.real_file_name)
                elif os.path.exists(self.tmp_file_name):
                    self.file_mime_type = mime.from_file(self.tmp_file_name)
                    stat = Path(self.tmp_file_name).stat()
            else:
                if os.path.exists(self.real_file_name):
                    self.file_mime_type = mime.from_file(self.real_file_name)
                    stat = Path(self.real_file_name).stat()
                elif os.path.exists(self.tmp_file_name):
                    self.file_mime_type = mime.from_file(self.tmp_file_name)
                    stat = Path(self.tmp_file_name).stat()

        if stat is not None and self.file_size is None:
            self.file_size = stat.st_size

        if self.stored_file_name is None:
            self.stored_file_name = str(uuid.uuid4()).upper()



# Generated by Django 3.2.7 on 2021-09-22 18:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('isc_common', '0133_auto_20210916_1226'),
    ]

    operations = [
        migrations.AlterField(
            model_name='text_informations',
            name='text',
            field=models.TextField(),
        ),
    ]



import logging

from django.db import transaction
from isc_common.http.DSResponse import JsonResponseWithException
from isc_common.models.upload_image import Common_UploadImage

logger = logging.getLogger(__name__)


@JsonResponseWithException()
class DSResponse_Fragment_params_UploadImage(Common_UploadImage):
    def __init__(self, request):
        from react.models.fragment_params import Fragment_params

        with transaction.atomic():
            file = request.FILES.get('upload_attatch')

            if file is not None:
                image, image_type, fragment_id = self.upload_image(request=request)
                res = Fragment_params.objects.filter(id=fragment_id).update(image=image)
                logger.debug(f'Updated: {res}')


#http://192.168.0.61:8003/logic/Imgs/Download/2608758?code=fragment_params&path=fragment_params&main_model=fragment_params&main_model_id=12&ws_host=176.107.243.22&ws_port=8002&ws_channel=common
#http://192.168.0.61:8003/logic/Imgs/Download/2607889?code=fragment_params&path=fragment_params&main_model=fragment_params&main_model_id=3&ws_host=176.107.243.22&ws_port=8002&ws_channel=common



import logging

from isc_common import Stack

logger = logging.getLogger(__name__)


class SSH_Client:
    def __init__(self, hostname, username, password, port=22):
        import paramiko
        import os
        import socket
        import traceback

        hostkey = None
        clients = None
        busy = False

        try:
            host_keys = paramiko.util.load_host_keys(
                os.path.expanduser("~/.ssh/known_hosts")
            )
        except IOError:
            try:
                # try ~/ssh/ too, because windows can't have a folder named ~/.ssh/
                host_keys = paramiko.util.load_host_keys(
                    os.path.expanduser("~/ssh/known_hosts")
                )
            except IOError:
                logger.debug("*** Unable to open host keys file")
                host_keys = {}

        if hostname in host_keys:
            hostkeytype = host_keys[hostname].keys()[0]
            hostkey = host_keys[hostname][hostkeytype]
            # logger.debug("Using host key of type %s" % hostkeytype)
            # logger.debug("Using host key %s" % str(hostkey))

        try:
            t = paramiko.Transport((hostname, port))
            t.connect(
                # hostkey=hostkey,
                username=username,
                password=password,
                gss_host=socket.getfqdn(hostname),
                gss_auth=False,
                gss_kex=False,
            )
            self.sftp = paramiko.SFTPClient.from_transport(t)
            self.hostname = hostname
            self.username = username
            self.password = password
            self.port = port
            self.busy = False
        except Exception as e:
            logger.error("*** Caught exception: %s: %s" % (e.__class__, e))
            traceback.print_exc()
            raise e

    def __str__(self):
        return f'-p {self.port} {self.username}@{self.hostname}'

    def reconnect(self):
        import paramiko
        import socket

        t = paramiko.Transport((self.hostname, self.port))
        t.connect(
            # hostkey=hostkey,
            username=self.username,
            password=self.password,
            gss_host=socket.getfqdn(self.hostname),
            gss_auth=False,
            gss_kex=False,
        )
        self.sftp = paramiko.SFTPClient.from_transport(t)
        self.clients.push(self)
        logger.debug(f'\nReconnect: {str(self)}')

    def get(self, remotepath, localpath, callback=None):
        try:
            self.busy = True
            self.sftp.get(remotepath=remotepath, localpath=localpath, callback=callback)
        except IOError as ex:
            logger.error(f'{remotepath} {str(ex)}')

    def put(self, localpath, remotepath, callback=None, confirm=True):
        try:
            self.sftp.put(localpath=localpath, remotepath=remotepath, callback=callback, confirm=confirm)
            return None
        except Exception as ex:
            logger.error(ex)
            return ex

    def getcwd(self):
        try:
            res = self.sftp.getcwd()
            return res, True
        except Exception as ex:
            logger.error(ex)
            return None, None

    def exists(self, path):
        try:
            self.sftp.stat(path)
            return True
        except FileNotFoundError:
            return False
        except IOError:
            return False
        except OSError:
            self.reconnect()
            self.sftp.stat(path)
            return True

    def listdir(self, path):
        return self.sftp.listdir(path)

    def getsize(self, path):
        if self.exists(path=path):
            return self.sftp.stat(path).st_size
        else:
            None

    def stat(self, path):
        if self.exists(path=path):
            return self.sftp.stat(path)
        else:
            None

    def open(self, filename, mode="r", bufsize=-1):
        if self.exists(path=filename):
            self.busy = True
            return self.sftp.open(filename=filename, mode=mode, bufsize=bufsize)
        else:
            None

    def remove(self, path):
        if self.exists(path=path):
            self.sftp.remove(path=path)
        else:
            None

    def rename(self, oldpath, newpath):
        self.sftp.rename(oldpath=oldpath, newpath=newpath)

    def chmod(self, path, mode):
        if self.exists(path=path):
            self.sftp.chmod(path=path, mode=mode)
        else:
            None

    def close(self):
        self.sftp.close()


class SSH_Clients:
    max_connect = 10;

    clients = Stack()

    def client(self, client_connect):
        # logger.debug(f'request connetc: {client_connect}')
        logger.debug(f'lenght stack: {self.clients.size()}')

        if self.clients.size() > self.max_connect:
            for connect in self.clients.stack[: - self.max_connect // 2]:
                connect.close()

            self.clients.stack = self.clients.stack[: -self.max_connect // 2]
            logger.debug(f'lenght stack: {self.clients.size()}')

        # clients = list(filter(lambda x: x.busy is False, self.clients.find(lambda client: client.hostname == client_connect.get('SSH_HOST') and
        #                                                                                   client.port == client_connect.get('SSH_PORT') and
        #                                                                                   client.username == client_connect.get('SSH_USER') and
        #                                                                                   client.password == client_connect.get('SSH_PASSWORD')
        #                                                                    )))
        # if len(clients) > 0:
        #     return clients[0]
        # else:
        res = SSH_Client(
            hostname=client_connect.get('SSH_HOST'),
            username=client_connect.get('SSH_USER'),
            password=client_connect.get('SSH_PASSWORD'),
            port=client_connect.get('SSH_PORT')
        )
        res.clients = self.clients
        self.clients.push(res)
        return res

    def close(self):
        for client in self.clients:
            client.close()
